/***************************************************
*	Module Name		:	Key_Board		   
*	Engineer			:	小梅哥
*	Target Device	:	10CL025YU256C8
*	Tool versions	:	Quartus Prime 18.1
*	Create Date		:	2021-12-20
*	Revision			:	v1.0
*	Description		:  矩阵键盘按键扫描检测
**************************************************/
module Key_Board(
	Clk,
	Rst_n,
	Key_Board_Row_i, //矩阵键盘行输入
	Key_Board_Col_o, //矩阵键盘列输出
	Key_Flag,
	Key_Value
);

	input Clk;
	input Rst_n;
	
	input [3:0]Key_Board_Row_i;
	output reg [3:0]Key_Board_Col_o;
	
	output reg Key_Flag;  //按键检查成功标志信号，每次检测成功输出一个时钟周期的高脉冲
	output reg [3:0]Key_Value; //键值
	
	reg [19:0] counter1;  //延时计数器
	reg En_Cnt;  //延时计数器使能控制信号，当进入延时消抖阶段时为高
	reg [10:0]state;  //状态寄存器，存储系统的状态
	reg [3:0]Key_Board_Row_r; //矩阵键盘行输入寄存器，存储行状态
	reg Cnt_Done;
	reg [3:0]Col_Tmp;
	reg [7:0]Key_Value_tmp;  //矩阵键盘扫描结果
	reg Key_Flag_r;   //按键检查成功标志信号
	
	localparam 
		IDLE				= 11'b00000000001,  //空闲态，无按键按下
		P_FILTER			= 11'b00000000010,  //按下消抖状态
		READ_ROW_P 		= 11'b00000000100,  //读取按下时矩阵键盘行状态
		SCAN_C0 			= 11'b00000001000,  //扫描矩阵键盘第0列（Col0）
		SCAN_C1 			= 11'b00000010000,  //扫描矩阵键盘第1列（Col1）
		SCAN_C2 			= 11'b00000100000,  //扫描矩阵键盘第2列（Col2）
		SCAN_C3 			= 11'b00001000000,  //扫描矩阵键盘第3列（Col3）
		PRESS_RESULT 	= 11'b00010000000,  //获得扫描结果
		WAIT_R			= 11'b00100000000,  //等待释放信号到来
		R_FILTER 		= 11'b01000000000,  //释放消抖
		READ_ROW_R 		= 11'b10000000000;  //读取释放时矩阵键盘行状态
		
//  延时计数器，计数以延时20ms
	always@(posedge Clk or negedge Rst_n)
	if(!Rst_n)
		counter1 <= 20'd0;
	else if(En_Cnt)begin
		if(counter1 == 20'd999999)
			counter1 <= 20'd0;
		else
			counter1 <= counter1 + 1'b1;
	end
	else
		counter1 <= 20'd0;

//产生延时完成标志信号，（当延时20ms时产生一个计数完成信号Cnt_Done）		
	always@(posedge Clk or negedge Rst_n)
	if(!Rst_n)
		Cnt_Done <= 1'b0;
	else if(counter1 == 20'd999999)
		Cnt_Done <= 1'b1;
	else
		Cnt_Done <= 1'b0;

//矩阵键盘扫描状态机  		
	always@(posedge Clk or negedge Rst_n)
	if(!Rst_n)begin
		En_Cnt <= 1'b0;
		state <= IDLE;
		Key_Board_Col_o <= 4'b0000;//默认需要让列输出为0，这样，当有按键按下时，行列接通，行输入才能读到有低电平
		Col_Tmp <= 4'd0;
		Key_Flag_r <= 1'b0;
		Key_Value_tmp <= 8'd0;
		Key_Board_Row_r <= 4'b1111;
	end
	else begin
		case(state)
			IDLE:   //空闲态，持续判断是否有按键被按下（按键按下时Key_Board_Row_i将不为全1）
				if(~&Key_Board_Row_i)begin
					En_Cnt <= 1'b1;
					state <= P_FILTER;			
				end
				else begin
					En_Cnt <= 1'b0;
					state <= IDLE;			
				end
				
			P_FILTER:   //按下消抖状态
				if(Cnt_Done) begin  //20ms延时完成
					En_Cnt <= 1'b0;    //停止延时计数器
					state <= READ_ROW_P;  //跳转到读取矩阵键盘行输入状态
				end
				else begin   //20ms延时还没有完成
					En_Cnt <= 1'b1;   //保持延时计数器继续计数
					state <= P_FILTER;				
				end
			
			READ_ROW_P:   //读取矩阵键盘行输入状态
				if(~Key_Board_Row_i)begin //行输入不全为1，表明确实有按键按下
					Key_Board_Row_r <= Key_Board_Row_i;  //将当期行输入状态存入行状态寄存器
					state <= SCAN_C0;	 //跳转到扫描矩阵键盘第0列（Col0）状态
					Key_Board_Col_o <= 4'b1110;  //将第0列列输出置0
				end
				else begin  //若行输入恢复为了全1，则表明此次为抖动，不再执行扫描，跳回空闲态
					state <= IDLE;
					Key_Board_Col_o <= 4'b0000;
				end
			
			SCAN_C0:   //扫描矩阵键盘第0列（Col0）状态
				begin
					state <= SCAN_C1;	 //下一个状态为扫描矩阵键盘第1列（Col1）状态
					Key_Board_Col_o <= 4'b1101;  //扫描第1列对应的行输出为1101
					if(~&Key_Board_Row_i)			//行输入不全为1，表明为当前列（第0列）有按键按下			
						Col_Tmp <= 4'b0001;     //将列值存入列寄存器中
					else
						Col_Tmp <= 4'b0000;
				end
				
			SCAN_C1:               //扫描矩阵键盘第1列（Col1）状态
				begin
					state <= SCAN_C2;	  //下一个状态为扫描矩阵键盘第2列（Col2）状态
					Key_Board_Col_o <= 4'b1011;  //扫描第1列对应的行输出为1011
					if(~&Key_Board_Row_i)				//行输入不全为1，表明为当前列（第1列）有按键按下		
						Col_Tmp <= Col_Tmp | 4'b0010;  //将列值存入列寄存器中
					else
						Col_Tmp <= Col_Tmp;				
				end
				
			SCAN_C2:    //扫描矩阵键盘第2列（Col2）状态
				begin
					state <= SCAN_C3;	  //下一个状态为扫描矩阵键盘第3列（Col3）状态
					Key_Board_Col_o <= 4'b0111;  //扫描第1列对应的行输出为0111
					if(~&Key_Board_Row_i)    //行输入不全为1，表明为当前列（第2列）有按键按下
						Col_Tmp <= Col_Tmp | 4'b0100;  //将列值存入列寄存器中
					else
						Col_Tmp <= Col_Tmp;				
				end
			
			SCAN_C3:   //扫描矩阵键盘第3列（Col3）状态
				begin
					state <= PRESS_RESULT;	//下一个状态为得到扫描结果状态
					if(~&Key_Board_Row_i)   //行输入不全为1，表明为当前列（第3列）有按键按下
						Col_Tmp <= Col_Tmp | 4'b1000;  //将列值存入列寄存器中
					else
						Col_Tmp <= Col_Tmp;				
				end
				
			PRESS_RESULT:  //得到扫描结果状态
				begin
					state <= WAIT_R;  //下一个状态为等待按键释放状态
					Key_Board_Col_o <= 4'b0000;  //让列输出全0，这样，当有按键按下时，行列接通，行输入才能读到有低电平
					
					
				/*4位行输入值相加为3，表明有且只有一个行输入为0，既保证只有一行中有按键被按下
         *4位列扫描结果相加为1，表明有且只有列中有按键被按下，通过这两个条件保证了一次
         *按键中只有一个按键被按下时才检测有效*/					
					if(((Key_Board_Row_r[0] + Key_Board_Row_r[1] + Key_Board_Row_r[2] + Key_Board_Row_r[3]) == 4'd3) && 
						((Col_Tmp[0] + Col_Tmp[1] + Col_Tmp[2] + Col_Tmp[3]) == 4'd1))begin
						Key_Flag_r <= 1'b1;  //产生检测成功标志信号
						Key_Value_tmp <= {Key_Board_Row_r,Col_Tmp};//将行列结果组合得到按键扫描结果
					end
					else begin  //若不满足只有一个按键被按下，则不产生检测成功标志信号
						Key_Flag_r <= 1'b0; //产生检测成功标志信号
						Key_Value_tmp <= Key_Value_tmp;
					end
				end
			WAIT_R:  //等待按键释放状态
				begin
					Key_Flag_r <= 1'b0; //清零检测成功标志信号
					if(&Key_Board_Row_i)begin  //无按键按下,表明按键被释放
						En_Cnt <= 1'b1;  //启动延时计数器
						state <= R_FILTER;  //进入释放消抖状态
					end
					else begin
						state <= WAIT_R;
						En_Cnt <= 1'b0;
					end
				end
			
			R_FILTER:			//释放消抖状态		
					if(Cnt_Done) begin  //延时时间到
						En_Cnt <= 1'b0;
						state <= READ_ROW_R;  //跳转入读取按键行输入状态（按键释放阶段）
					end
					else begin
						En_Cnt <= 1'b1;
						state <= R_FILTER;				
					end
				
			READ_ROW_R:  //读取按键行输入状态（按键释放阶段）
				if(&Key_Board_Row_i)  //无按键按下，表明确实已经稳定的释放了
					state <= IDLE;
				else begin
					En_Cnt <= 1'b1;   //跳转回空闲状态
					state <= R_FILTER;		//否则表明此次释放为抖动，回到释放消抖状态继续等待		
				end
			default:state <= IDLE;
		endcase	
	end
	
	//将扫描结果译码输出   	
	always@(posedge Clk or negedge Rst_n)
	if(!Rst_n)begin
		Key_Value <= 4'd0;
	end
	else if(Key_Flag_r)begin      //按键检查成功输出标志
		case(Key_Value_tmp)
			8'b1110_0001 : Key_Value <= 4'd0;  //第0行，第0列
			8'b1110_0010 : Key_Value <= 4'd1;  //第0行，第1列
			8'b1110_0100 : Key_Value <= 4'd2;  //第0行，第2列
			8'b1110_1000 : Key_Value <= 4'd3;  //第0行，第3列
			
			8'b1101_0001 : Key_Value <= 4'd4;  //第1行，第0列
			8'b1101_0010 : Key_Value <= 4'd5;  //第1行，第1列
			8'b1101_0100 : Key_Value <= 4'd6;  //第1行，第2列
			8'b1101_1000 : Key_Value <= 4'd7;  //第1行，第3列
			
			8'b1011_0001 : Key_Value <= 4'd8;  //第2行，第0列
			8'b1011_0010 : Key_Value <= 4'd9;  //第2行，第1列
			8'b1011_0100 : Key_Value <= 4'd10;  //第2行，第2列
			8'b1011_1000 : Key_Value <= 4'd11;  //第2行，第3列
			
			8'b0111_0001 : Key_Value <= 4'd12;  //第3行，第0列
			8'b0111_0010 : Key_Value <= 4'd13;  //第3行，第1列
			8'b0111_0100 : Key_Value <= 4'd14;  //第3行，第2列
			8'b0111_1000 : Key_Value <= 4'd15;  //第3行，第3列
			default:begin Key_Value <= Key_Value;end
		endcase
	end
	
	always@(posedge Clk)
		Key_Flag <= Key_Flag_r;
		
endmodule
